#include "cifrasProduto.hpp"

uint16_t CifrasProduto::cifraProduto (uint16_t msg,uint16_t chave) {
  uint16_t aux1,aux2,aux3,aux4;
  uint16_t cfr1,cfr2;
  
  // E_k(m) = m XOR k
  cfr1 = msg^chave; // m1m2m3m4 m5m6m7m8 m9m10m11m12 m13m14m15m16
  
  // Permutação m1--m4 m5--m8 m9--m12 m13--m16 ---> m13--m16 m9--m12 m5--m8 m1--m4

  // Separar as secções correspondentes com máscaras apropriadas
  // Mover para a posição final

  aux1 = cfr1 & 0xF000; // 1111 0000 0000 0000 --> m1--m4  0000   0000   0000
  aux1 = aux1>>12;      //                     -->  0000   0000   0000  m1--m4
  
  aux2 = cfr1 & 0x0F00; // 0000 1111 0000 0000 -->  0000  m5--m8  0000   0000  
  aux2 = aux2>>4;       //                     -->  0000   0000  m5--m8  0000
  
  aux3 = cfr1 & 0x00F0; // 0000 0000 1111 0000 -->  0000   0000  m9--m12 0000
  aux3 = aux3<<4;       //        s             -->  0000  m9--m12 0000   0000
  
  aux4 = cfr1 & 0x000F; // 0000 0000 0000 1111 -->  0000   0000   0000  m13--m16
  aux4 = aux4<<12;      //                     -->m13--m16 0000   0000   0000

  // Juntar as diferentes secções no resultado final
  cfr2 = aux1 | aux2 | aux3 | aux4;

  return(cfr2);
  
}


void CifrasProduto::encriptarCifraProduto(FILE *pfM, FILE *pfC, uint16_t chave){
  uint16_t aux1,aux2,aux3;
  int contador=0;
  int carLido,carEscrever,cifrado;
  
  /*
   * getc: On success, the character read is returned (promoted to an int value).
   * int - 4 bytes
   * putc: The value is internally converted to an unsigned char when written.
   */

  // A chave tem comprimento 16bits 
  
  while ((carLido = getc(pfM)) != EOF)  {   // ler 1 char --> int (4bytes)
    if (contador == 0) { // se o contador=0, primeiro de dois char, guarda-se em aux1
      contador = 1; // incrementa-se o contador
      aux1 = carLido;        
    }
    else { // se o contador=1, segundo dos dois chars 
      contador = 0; // re-inicializa-se o contador
      aux1 = aux1 << 8; // aux1 = x1x2x3x4x5x6x7x8 00000000
      aux2 = carLido;   // aux2 = 00000000 x9x10x11x12x13x14x15x16
      aux3 = aux1 | aux2;  // aux3 = x1x2...x16
      cifrado = cifraProduto(aux3,chave); // c = y1y2...y16
      carEscrever = (cifrado>>8);  // 000000000 y1y2...y8
      putc(carEscrever,pfC); 
      cifrado = (cifrado<<8); // y9y10...y16 00000000
      carEscrever = (cifrado>>8); // 00000000 y9y10...y16
      putc(carEscrever,pfC);
    }
  }
  
  /*
   * Se foi feito primeiro a desencriptação, o ficheiro já está alinhado,
   * como tal esta parte será ignorada
   */
  if (contador == 1) { // se a mensagem tem 1 char a mais 
    cifrado = cifraProduto(aux1,chave); // encripta o último char lido (8bits) 
    carEscrever = (cifrado>>8);  // alinha para a escrita
    putc(carEscrever,pfC);
    putc(carEscrever,pfC); // para alinhar, repete o útimo carácter
  }
}


uint16_t CifrasProduto::decifraProduto(uint16_t cfr,uint16_t chave){
  uint16_t aux1,aux2,aux3,aux4;
  uint16_t cfr1,cfr2;

  // Inverso da cifra
  
  // Permutação <-- m13--m16 m9--m12 m5--m8   m1--m4
  //            -->  m1--m4  m5--m8  m9--m12 m13--m16

  aux1 = cfr & 0x000F; // 0000 0000 0000 1111 -->  0000   0000   0000  m1--m4
  aux1 = aux1<<12;   //                     --> m1--m4  0000   0000   0000
  
  aux2 = cfr & 0x00F0; // 0000 0000 1111 0000 -->  0000   0000  m5--m8  0000
  aux2 = aux2<<4;    //                     -->  0000  m5--m8  0000   0000  
  
  aux3 = cfr & 0x0F00; // 0000 1111 0000 0000 -->  0000  m9--m12 0000   0000
  aux3 = aux3>>4;    //                     -->  0000   0000  m9--m12 0000
  
  aux4 = cfr & 0xF000; // 1111 0000 0000 0000 -->m13--m16 0000   0000   0000
  aux4 = aux4>>12;   //                     -->  0000   0000   0000  m13--m16

  // Juntar as diferentes secções no resultado final
  cfr1 = aux1 | aux2 | aux3 | aux4;  
  
  // c XOR chave
  cfr2 = cfr1 ^ chave;
  
  return(cfr2);
  
}

void CifrasProduto::desencriptarProduto (FILE *pfM, FILE *pfC, uint16_t chave) {
  uint16_t c;
  char letra;
  int contador=0,x;

  uint16_t aux1,aux2,aux3;
  int carLido,carEscrever,decifrado;

  
  while ((carLido = getc(pfM)) != EOF)  {   
    if (contador == 0) {
      contador = 1;
      aux1 = carLido;      
    }
    else {
      contador = 0;
      aux1 = aux1<<8;
      aux2 = carLido;
      aux3 = aux1 | aux2;
      decifrado = decifraProduto(aux3,chave);
      carEscrever = (decifrado>>8);

      putc(carEscrever,pfC);
      
      decifrado = decifrado<<8;
      carEscrever = decifrado>>8;

      putc(carEscrever,pfC);
    }
  }

  /*
   * Se foi feito primeiro a encriptação, o ficheiro já está alinhado,
   * como tal esta parte será ignorada
   */
  if (contador == 1) {
    decifrado = decifraProduto (aux1,chave);
    carEscrever = (decifrado>>8);

    putc(carEscrever,pfC);
    putc(carEscrever,pfC); // o último carácter é repetido
  }
}
